# !!!! WORK IN PROGRESSS !!!!
#
# TEAM benchmark 23
# https://www.compumag.org/jsite/team.html
#
# !!!! WORK IN PROGRESSS !!!!


from ngsolve import *
from netgen.csg import *

if True:
    # Configuration A, large
    d1 = 0.003048
    d2 = 0.0039624
    d3 = 0.0029972
    l = 0.0016
    lc = 0.001524
    sc = 0.000127
    delta = 0.000254
else:
    # Configuration B, small
    d1 = 0.001524
    d2 = 0.003175
    d3 = 0.0016
    l = 0.000818
    lc = 0.001524    
    sc = 0.000127
    delta = 0.000254


geom = CSGeometry()
geom.SetBoundingBox(Pnt(-0.1,-0.1,-0.1), Pnt(0.1,0.1,0.1))
magnet = (Cylinder(Pnt(delta,0,0), Pnt(delta+l,0,0), d3/2) * Plane(Pnt(delta,0,0), Vec(-1,0,0)) * Plane(Pnt(delta+l,0,0), Vec(1,0,0))).bc("gamma")
coil =  ((Cylinder(Pnt(-lc-sc,0,0), Pnt(-sc,0,0), d2/2)-Cylinder(Pnt(-lc-sc,0,0), Pnt(-sc,0,0), d1/2)) * Plane(Pnt(-lc-sc,0,0), Vec(-1,0,0)) * Plane(Pnt(-sc,0,0), Vec(1,0,0))).bc("gammacoil")
r = 0.01
box = OrthoBrick (Pnt(-r,-r,-r), Pnt(r,r,r))

geom.Add (magnet.mat("magnet").maxh(0.001))
geom.Add (coil.mat("coil").maxh(0.001))
geom.Add ( (box-magnet-coil).mat("air"))

mesh = Mesh(geom.GenerateMesh(maxh=0.01))
mesh.Curve(3)

fes = HCurl(mesh, order=3, nograds=True)
print ("ndof = ", fes.ndof)
u,v = fes.TnT()

with TaskManager():
    mu0 = 1.257e-6
    mur = { "magnet" : 100, "air" : 1, "coil" : 1 }
    nu = 1.0 / mu0 / CoefficientFunction( [mur[mat] for mat in mesh.GetMaterials()] )
    a = BilinearForm(fes)
    a += SymbolicBFI (nu*curl(u)*curl(v) + nu*1e-6*u*v)

    c = Preconditioner(a, "bddc")
    a.Assemble()

    f = LinearForm(fes)
    tau = CoefficientFunction((0, z, -y))
    tau = 1.0/Norm(tau)*tau
    curdens = 0.1 * 280  / (lc*(d2-d1)/2) # Amper turns / coil area
    f += SymbolicLFI(curdens*tau*v, definedon=mesh.Materials("coil"))
    f.Assemble()
    
    gfu = GridFunction(fes)
    solvers.CG(mat=a.mat, pre=c.mat, rhs=f.vec, sol=gfu.vec)


B = curl(gfu)


# Maxwell stress tensor ...
sigma = nu * (OuterProduct(B, B) - 0.5 * (B*B) * Id(3) )


# compute force in H1-space:
# compute f = div sigma  in weak form
fes = VectorH1(mesh, order=3)
u,v = fes.TnT()
force = LinearForm(fes)
force += SymbolicLFI( InnerProduct(sigma, grad(v)))
force.Assemble()

mass = BilinearForm(fes)
mass += SymbolicBFI(u*v, BND)
mass.Assemble()

gfforce = GridFunction(fes)
gfforce.vec.data = mass.mat.Inverse(inverse="sparsecholesky", freedofs=fes.GetDofs(mesh.Boundaries("gamma|gammacoil"))) * force.vec


# compute surface force in Facet-space as  [sigma n]
fes1 = FacetFESpace(mesh, order=3)
fes = FESpace( [fes1, fes1, fes1] )
u = fes.TrialFunction()
v = fes.TestFunction()
uvec = CoefficientFunction( (u[0], u[1], u[2]) )
vvec = CoefficientFunction( (v[0], v[1], v[2]) )
n = specialcf.normal(3)

force = LinearForm(fes)
force += SymbolicLFI( (sigma*n)*vvec, element_boundary=True)  # produces jump-terms
force.Assemble()

mass = BilinearForm(fes)
mass += SymbolicBFI(0.5*uvec*vvec, element_boundary=True)  # facet mass matrix (1/2 from both sides)
mass.Assemble()

gffacetforce = GridFunction(fes)
gffacetforce.vec.data = mass.mat.Inverse(inverse="sparsecholesky", freedofs=fes.GetDofs(mesh.Boundaries("gamma|gammacoil")) ) * force.vec
facetforcevec = CoefficientFunction( (gffacetforce.components[0], gffacetforce.components[1], gffacetforce.components[2]) )

char_coil = CoefficientFunction( [1 if bc=="gammacoil" else 0 for bc in mesh.GetBoundaries()] )
char_core = CoefficientFunction( [1 if bc=="gamma" else 0 for bc in mesh.GetBoundaries()] )

print ("facetforce-h1, core: ", Integrate(gfforce*char_core, mesh, BND))
print ("facetforce-facet, core: ", Integrate(facetforcevec*char_core, mesh, BND))

print ("facetforce-h1, coil: ", Integrate(gfforce*char_coil, mesh, BND))
print ("facetforce-facet, coil: ", Integrate(facetforcevec*char_coil, mesh, BND))


Draw (curl(gfu), mesh, "B-field")
Draw (gfforce, mesh, "facetforce-H1")
Draw (facetforcevec, mesh, "facetforce", draw_vol=False)


